from blankly.enums import Side
from blankly.exchanges.interfaces.futures_exchange_interface import FuturesExchangeInterface
from blankly.frameworks.strategy import FuturesStrategyState


def close_position(symbol: str, state: FuturesStrategyState):
    """
    Exit a position
    Args:
        state: the StrategyState
        symbol: the symbol to sell
    """
    position = state.interface.get_position(symbol)
    if not position:
        return

    if position['size'] < 0:
        side = Side.BUY
    elif position['size'] > 0:
        side = Side.SELL
    else:
        # wtf?
        return

    state.interface.market_order(symbol, side, abs(position['size']), reduce_only=True)



"""
    binance ticker class.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import traceback

import blankly
import blankly.exchanges.interfaces.binance.binance_websocket_utils as websocket_utils
from blankly.exchanges.interfaces.websocket import Websocket
from blankly.utils.utils import info_print


class Tickers(Websocket):
    def __init__(self, symbol, stream, log=None, initially_stopped=False,
                 websocket_url="wss://stream.binance.{}:9443/ws", **kwargs):
        """
        Create and initialize the ticker
        Args:
            symbol: Currency to initialize on such as "btcusdt"
            stream: Stream to use, such as "depth" or "trade"
            log: Fill this with a path to a log file that should be created
            websocket_url: Default websocket URL feed.
        """
        # Reload preferences
        self.__preferences = blankly.utils.load_user_preferences()

        self.__logging_callback, self.__interface_callback, log_message = websocket_utils.switch_type(stream)
        url = websocket_url.format(self.__preferences['settings']['binance']['binance_tld'])

        super().__init__(symbol, stream, log, log_message, url, None, kwargs)

        # Start the websocket
        if not initially_stopped:
            self.start_websocket(
                self.on_open,
                self.on_message,
                self.on_error,
                self.on_close,
                self.read_websocket
            )

    def read_websocket(self):
        # Main thread to sit here and run
        self.ws.run_forever()
        # This repeats the close behavior just in case something happens

    def on_message(self, ws, message):
        """
        Exchange specific actions to perform when receiving a message
        """
        self.message_count += 1
        message = json.loads(message)
        try:
            self.most_recent_time = message['E']
            self.time_feed.append(self.most_recent_time)

            self.log_response(self.__logging_callback, message)

            interface_message = self.__interface_callback(message)
            self.ticker_feed.append(interface_message)
            self.most_recent_tick = interface_message
            for i in self.callbacks:
                i(interface_message, **self.kwargs)
        except KeyError:
            # If the try below figures this out then we don't have to traceback
            error_found = False
            try:
                if message['result'] is None:
                    self.response = message
                    error_found = True
            except KeyError:
                traceback.print_exc()

            if not error_found:
                traceback.print_exc()

    def on_error(self, ws, error):
        info_print(error)

    def on_close(self, ws):
        # This repeats the close behavior just in case something happens
        pass

    def on_open(self, ws):
        request = json.dumps({
            'method': 'SUBSCRIBE',
            'params': [f'{self.symbol}@{self.stream}'],
            'id': 1
        })
        ws.send(request)

    def restart_ticker(self):
        self.start_websocket(
            self.on_open,
            self.on_message,
            self.on_error,
            self.on_close,
            self.read_websocket
        )



"""
    binance exchange definitions and setup
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from binance.client import Client

from blankly.exchanges.exchange import Exchange
from blankly.exchanges.auth.auth_constructor import AuthConstructor
from blankly.utils import utils


class Binance(Exchange):
    def __init__(self, portfolio_name=None, keys_path="keys.json", settings_path=None):
        Exchange.__init__(self, "binance", portfolio_name, settings_path)

        # Load the auth from the keys file
        auth = AuthConstructor(keys_path, portfolio_name, 'binance', ['API_KEY', 'API_SECRET', 'sandbox'])

        sandbox = super().evaluate_sandbox(auth)

        if sandbox:
            calls = Client(api_key=auth.keys['API_KEY'], api_secret=auth.keys['API_SECRET'],
                           tld=self.preferences["settings"]['binance']["binance_tld"],
                           testnet=True)
        else:
            calls = Client(api_key=auth.keys['API_KEY'], api_secret=auth.keys['API_SECRET'],
                           tld=self.preferences["settings"]['binance']["binance_tld"])

        # Always finish the method with this function
        super().construct_interface_and_cache(calls)

    """
    Builds information about the symbol on this exchange by making particular API calls
    """

    def get_asset_state(self, symbol):
        """
        Portfolio state is the internal properties for the exchange block.
        """
        # TODO Populate this with useful information
        symbol = utils.get_base_asset(symbol)
        account = self.interface.get_account(symbol=symbol)
        return account

    def get_exchange_state(self):
        """
        Exchange state is the external properties for the exchange block
        """
        # TODO Populate this with useful information
        return self.interface.get_products()

    def get_direct_calls(self) -> Client:
        return self.calls



"""
    Class for creating general websocket streams
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import random

import requests

import blankly.utils.utils
from blankly.exchanges.interfaces.alpaca.alpaca_websocket import Tickers as Alpaca_Websocket
from blankly.exchanges.interfaces.binance.binance_websocket import Tickers as Binance_Websocket
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro_websocket import Tickers as Coinbase_Pro_Websocket
from blankly.exchanges.interfaces.kucoin.kucoin_websocket import Tickers as Kucoin_Websocket
from blankly.exchanges.interfaces.ftx.ftx_websocket import Tickers as Ftx_Websocket
from blankly.exchanges.interfaces.okx.okx_websocket import Tickers as Okx_Websocket
from blankly.exchanges.managers.websocket_manager import WebsocketManager


class GeneralManager(WebsocketManager):
    def __init__(self, default_exchange, default_symbol):
        self.__default_exchange = default_exchange
        self.__default_currency = default_symbol

        self.__websockets = {}

        super().__init__(self.__websockets, default_symbol, default_exchange)

    def create_general_connection(self, callback, channel, log=None, override_symbol=None, override_exchange=None):
        """
        Create a channel on any given stream on any exchange. This is not stored in the manager object, only returned.
        """
        asset_id_cache = self.__default_currency
        exchange_cache = self.__default_exchange

        if override_symbol is not None:
            asset_id_cache = override_symbol

        if override_exchange is not None:
            exchange_cache = override_exchange

        if channel not in self.__websockets.keys():
            self.__websockets[channel] = {}

        if exchange_cache not in self.__websockets[channel].keys():
            self.__websockets[channel][exchange_cache] = {}

        if asset_id_cache not in self.__websockets[channel][exchange_cache].keys():
            self.__websockets[channel][exchange_cache][asset_id_cache] = {}

        use_sandbox = self.preferences['settings']['use_sandbox_websockets']

        if exchange_cache == "coinbase_pro":
            if use_sandbox:
                websocket = Coinbase_Pro_Websocket(asset_id_cache, channel, log,
                                                   WEBSOCKET_URL="wss://ws-feed-public.sandbox.pro.coinbase.com")
            else:
                websocket = Coinbase_Pro_Websocket(asset_id_cache, channel, log)
            websocket.append_callback(callback)

            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket

            return websocket
        elif exchange_cache == "kucoin":
            if use_sandbox:
                response = requests.post('https://trade-sandbox.kucoin.com/_api/bullet-usercenter/v1/bullet-public'
                                         ).json()

                base_endpoint = response['data']['instanceServers'][0]['endpoint']
                token = response['data']['token']
                websocket = Kucoin_Websocket(asset_id_cache, channel, f"{base_endpoint}/socket.io/?token={token}", log)
                # wss://push-socketio-sandbox.kucoin.com
            else:
                response = requests.post('https://api.kucoin.com/api/v1/bullet-public').json()

                base_endpoint = response['data']['instanceServers'][0]['endpoint']
                token = response['data']['token']

                websocket = Kucoin_Websocket(asset_id_cache, channel,
                                             f"{base_endpoint}/api/v1/?token={token}&[connectId="
                                             f"{random.randint(1, 100000000) * 100000000}]", log)
            websocket.append_callback(callback)

            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket

            return websocket
        elif exchange_cache == "ftx":
            if use_sandbox:
                raise ValueError("Error: FTX does not have a sandbox mode")
            else:
                websocket = Ftx_Websocket(override_symbol, asset_id_cache, channel, log)

            websocket.append_callback(callback)

            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket

            return websocket
        elif exchange_cache == "okx":
            if use_sandbox:
                websocket = Okx_Websocket(asset_id_cache, channel, log,
                                          WEBSOCKET_URL="wss://wspap.okx.com:8443/ws/v5/public?brokerId=9999")
            else:
                websocket = Okx_Websocket(asset_id_cache, channel, log)
            websocket.append_callback(callback)

            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket
        elif exchange_cache == "binance":
            # Lower this to subscribe
            asset_id_cache = blankly.utils.to_exchange_symbol(asset_id_cache, "binance").lower()
            if use_sandbox:
                websocket = Binance_Websocket(asset_id_cache, channel, log,
                                              WEBSOCKET_URL="wss://testnet.binance.vision/ws")
            else:
                websocket = Binance_Websocket(asset_id_cache, channel, log)
            websocket.append_callback(callback)

            # Upper this to cache
            asset_id_cache = asset_id_cache.upper()
            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket

            return websocket

        elif exchange_cache == "alpaca":
            stream = self.preferences['settings']['alpaca']['websocket_stream']

            asset_id_cache = blankly.utils.to_exchange_symbol(asset_id_cache, "alpaca")
            if use_sandbox:
                websocket = Alpaca_Websocket(asset_id_cache, channel, log,
                                             WEBSOCKET_URL=
                                             "wss://paper-api.alpaca.markets/stream/v2/{}/".format(stream))
            else:
                websocket = Alpaca_Websocket(asset_id_cache, channel, log,
                                             WEBSOCKET_URL="wss://stream.data.alpaca.markets/v2/{}/".format(stream))
            websocket.append_callback(callback)

            self.__websockets[channel][exchange_cache][asset_id_cache] = websocket

            return websocket

    """
    Create overriden method signatures
    """

    def append_callback(self, callback_object, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        super().append_callback(callback_object, override_symbol, override_exchange)

    def is_websocket_open(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().is_websocket_open(override_symbol, override_exchange)

    def get_most_recent_time(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().get_most_recent_time(override_symbol, override_exchange)

    def get_time_feed(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().get_time_feed(override_symbol, override_exchange)

    def get_feed(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().get_feed(override_symbol, override_exchange)

    def get_response(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().get_response(override_symbol, override_exchange)

    def close_websocket(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().close_websocket(override_symbol, override_exchange)

    def restart_ticker(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().restart_ticker(override_symbol, override_exchange)

    def get_most_recent_tick(self, channel, override_symbol=None, override_exchange=None):
        self.websockets = self.__websockets[channel]
        return super().get_most_recent_tick(override_symbol, override_exchange)



"""
    __init__ file to give the module access to the libraries.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import blankly.utils.utils
import blankly.data as data
from blankly.exchanges.interfaces.coinbase_pro.coinbase_pro import CoinbasePro
from blankly.exchanges.interfaces.binance.binance import Binance
from blankly.exchanges.interfaces.alpaca.alpaca import Alpaca
from blankly.exchanges.interfaces.oanda.oanda import Oanda
from blankly.exchanges.interfaces.kucoin.kucoin import Kucoin
from blankly.exchanges.interfaces.ftx.ftx import FTX
from blankly.exchanges.interfaces.okx.okx import Okx
from blankly.exchanges.interfaces.paper_trade.paper_trade import PaperTrade
from blankly.exchanges.interfaces.keyless.keyless import KeylessExchange
from blankly.frameworks.strategy import Strategy as Strategy
from blankly.frameworks.model.model import Model as Model
from blankly.frameworks.strategy import StrategyState as StrategyState
from blankly.frameworks.screener.screener import Screener
from blankly.frameworks.screener.screener_state import ScreenerState

from blankly.exchanges.managers.ticker_manager import TickerManager
from blankly.exchanges.managers.orderbook_manager import OrderbookManager
from blankly.exchanges.managers.general_stream_manager import GeneralManager
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface as Interface
from blankly.frameworks.multiprocessing.blankly_bot import BlanklyBot
from blankly.utils.utils import trunc
import blankly.utils.utils as utils
from blankly.utils.scheduler import Scheduler
import blankly.indicators as indicators
from blankly.utils import time_builder

from blankly.enums import Side, OrderType, OrderStatus, TimeInForce
from blankly.exchanges.interfaces.binance_futures.binance_futures import BinanceFutures
from blankly.exchanges.interfaces.ftx_futures.ftx_futures import FTXFutures
from blankly.frameworks.strategy import FuturesStrategy
from blankly.frameworks.strategy import FuturesStrategyState

from blankly.deployment.reporter_headers import Reporter as __Reporter_Headers

is_deployed = False
_screener_runner = None

_backtesting = blankly.utils.check_backtesting()
try:
    from blankly_external import Reporter as __Reporter

    reporter = __Reporter
    is_deployed = True
except ImportError:
    reporter = __Reporter_Headers()



"""
    Tulipy wrappers
    Copyright (C) 2021 Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import tulipy as ti

from blankly.indicators.utils import check_series, convert_to_numpy


def bbands(data, period=14, stddev=2):
    data = convert_to_numpy(data)
    return ti.bbands(data, period, stddev)


def wad(high_data, low_data, close_data, use_series=False):
    if check_series(high_data) or check_series(low_data) or check_series(close_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    close_data = convert_to_numpy(close_data)
    wad = ti.wad(high_data, low_data, close_data)
    return pd.Series(wad) if use_series else wad


def wilders(data, period=50, use_series=False):
    if check_series(data):
        use_series = True
    data = convert_to_numpy(data)
    wilders = ti.wilders(data, period)
    return pd.Series(wilders) if use_series else wilders


def willr(high_data, low_data, close_data, period=50, use_series=False):
    if check_series(high_data) or check_series(low_data) or check_series(close_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    close_data = convert_to_numpy(close_data)
    willr = ti.willr(high_data, low_data, close_data, period)
    return pd.Series(willr) if use_series else willr


def true_range(high_data, low_data, close_data, use_series=False):
    if check_series(high_data) or check_series(low_data) or check_series(close_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    close_data = convert_to_numpy(close_data)
    tr = ti.tr(high_data, low_data, close_data)
    return pd.Series(tr) if use_series else tr


def average_true_range(high_data, low_data, close_data, period=50, use_series=False):
    if check_series(high_data) or check_series(low_data) or check_series(close_data):
        use_series = True
    high_data = convert_to_numpy(high_data)
    low_data = convert_to_numpy(low_data)
    close_data = convert_to_numpy(close_data)
    atr = ti.atr(high_data, low_data, close_data, period=period)
    return pd.Series(atr) if use_series else atr



"""
    Class to carry out orders
    Copyright (C) 2021  Emerson Dove, Brandon Fan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


class Order:
    """Order Type"""

    def __init__(self, order_type, side, amount, price=None):
        if order_type not in ['market', 'limit']:
            raise ValueError(
                "Order types can only be of type ['market', 'limit'], but was given: {}".format(order_type))
        self.type = order_type
        if side not in ['buy', 'sell']:
            raise ValueError("Order side can only be of type ['buy', 'sell'], but was given: {}".format(order_type))
        self.side = side
        if self.amount <= 0:
            raise ValueError("Size must be greater than zero, but was given: {}".format(size))
        self.amount = amount
        if self.type == 'limit' and price is None:
            raise ValueError("Limit orders require a price but None was given")

        if price <= 0:
            raise ValueError("Price must be greater than zero but was given: {}".format(price))

        self.price = price



"""
    Paper trade exchange constructed from other interfaces
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import threading
import time
import traceback
import warnings

import blankly.exchanges.interfaces.paper_trade.utils as paper_trade
import blankly.utils.utils as utils
from blankly.exchanges.interfaces.paper_trade.local_account.trade_local import LocalAccount
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface
from blankly.exchanges.interfaces.exchange_interface import ExchangeInterface
from blankly.exchanges.interfaces.paper_trade.backtesting_wrapper import BacktestingWrapper
from blankly.exchanges.orders.limit_order import LimitOrder
from blankly.exchanges.orders.market_order import MarketOrder
from blankly.exchanges.orders.stop_loss import StopLossOrder
from blankly.exchanges.orders.take_profit import TakeProfitOrder
from blankly.utils.exceptions import APIException, InvalidOrder


class PaperTradeInterface(ExchangeInterface, BacktestingWrapper):
    def __init__(self, derived_interface: ABCExchangeInterface, initial_account_values: dict = None):
        # This paper trade orders keeps a live track of the orders
        self.paper_trade_orders = []
        # These two keep track of which limit orders and when the order finishes
        self.canceled_orders = []
        self.executed_orders = []

        # Save the market order execution details for later
        self.market_order_execution_details = []

        self.get_products_cache = None
        self.get_fees_cache = {}
        self.get_order_filter_cache = {}

        self.__run_watchdog = True

        self.__thread = None

        # Set the type of the paper trade exchange to be the same as whichever interface its derived from
        ExchangeInterface.__init__(self, derived_interface.get_exchange_type(), derived_interface)
        BacktestingWrapper.__init__(self)

        self.__local_account_cache = None
        self.__initial_account_values = initial_account_values

        # Initialize our traded assets list
        self.traded_assets = []

        # self.evaluate_traded_account_assets()

        self.__enable_shorting = self.user_preferences['settings']['alpaca']['enable_shorting']
        self.__calculate_margin = self.user_preferences['settings']['simulate_margin']

        # This logically overrides any __enable_shorting
        self.__force_shorting = self.user_preferences['settings']['global_shorting']

        if self.user_preferences['settings']['paper']['price_source'] == 'websocket':
            from blankly.exchanges.managers.ticker_manager import TickerManager
            warnings.warn("Experimental websocket prices enabled.")
            self.__ticker_manager = TickerManager(self.get_exchange_type(), default_symbol='')
            self._websocket_update = lambda *args: None

    @property
    def local_account(self):
        if self.__local_account_cache is None:
            if self.get_exchange_type() == 'keyless':
                accounts = utils.add_all_products({}, self.get_products())
            else:
                # Write in the accounts to our local account. This involves getting the values directly from the
                # exchange
                accounts = self.calls.get_account()
            # If it's given an initial account value dictionary, write all the account values that we passed in to the
            # function to the account, then default everything else to zero.
            if self.__initial_account_values is not None:
                for i in accounts.keys():
                    if i in self.__initial_account_values.keys():
                        accounts[i] = utils.AttributeDict({
                            'available': self.__initial_account_values[i],
                            'hold': 0.0
                        })
                    else:
                        accounts[i] = utils.AttributeDict({
                            'available': 0.0,
                            'hold': 0.0
                        })

            # Initialize the local account
            self.__local_account_cache = LocalAccount(accounts)
            return self.__local_account_cache
        else:
            return self.__local_account_cache

    def evaluate_traded_account_assets(self):
        # Because alpaca has so many columns we need to optimize to perform an accurate backtest
        accounts = self.get_account()

        for i in accounts.keys():
            if (accounts[i]['available'] + accounts[i]['hold']) != 0 and i not in self.traded_assets:
                self.traded_assets.append(i)

    def backtesting_time(self):
        """
        Return the backtest time if we're backtesting, if not just return None. If it's None the caller
         assumes no backtesting. The function being overridden always returns None
        """
        # This is for the inits because it happens with both live calls and in the past
        if self.initial_time is not None and not self.backtesting:
            return self.initial_time
        # This is for the actual price loops
        if self.backtesting:
            return self.time()
        else:
            return None

    def __check_trading_assets(self, symbol: str):
        """
        Append a newly traded symbol to traded assets array. This is used to more efficiently evaluate
        which accounts need graphs

        Args:
            symbol (str): The string representing the asset
        """
        base_asset = utils.get_base_asset(symbol)
        quote_asset = utils.get_quote_asset(symbol)

        if base_asset not in self.traded_assets:
            self.traded_assets.append(base_asset)

        if quote_asset not in self.traded_assets:
            self.traded_assets.append(quote_asset)

    def init_exchange(self):
        try:
            self.calls.get_exchange_type()
        except AttributeError:
            traceback.print_exc()
            raise AttributeError("Are you passing a non-exchange object into the paper trade constructor?")

    """ Needs to be overridden here """

    def start_paper_trade_watchdog(self):
        # TODO, this process could use variable update time/websocket usage, poll `time` and a variety of settings
        #  to create a robust trading system
        # Create the watchdog for watching limit orders
        self.__thread = threading.Thread(target=self.__paper_trade_watchdog, daemon=True)
        self.__thread.start()
        self.__run_watchdog = True

    def stop_paper_trade_watchdog(self):
        self.__run_watchdog = False

    """ Needs to be overridden here """

    def __paper_trade_watchdog(self):
        """
        Internal order watching system
        """
        utils.info_print('Evaluating paper limit orders every 10 seconds...')
        while True:
            time.sleep(10)
            if not self.__run_watchdog:
                break
            self.evaluate_limits()

    @staticmethod
    def __get_decimals(number) -> int:
        return utils.count_decimals(number)

    def override_local_account(self, value_dictionary: dict):
        """
        Push a new set of initial account values to the algorithm. All values not given in the
        value dictionary that currently exist will be set to zero.

        Args:
            value_dictionary (dict): Account dictionary of format {'BTC': 2.3, 'GRT': 1.1}
        """
        current_account = self.local_account.get_accounts()
        for k, v in current_account.items():
            if k in value_dictionary.keys():
                current_account[k] = utils.AttributeDict({
                    'available': value_dictionary[k],
                    'hold': 0
                })
            else:
                current_account[k] = utils.AttributeDict({
                    'available': 0,
                    'hold': 0
                })
        self.local_account.override_local_account(current_account)

    def evaluate_limits(self):
        """
        When this is run it checks the local paper trade orders to see if any need to go through
        """
        used_currencies = []
        for i in self.paper_trade_orders:
            if i["symbol"] not in used_currencies:
                used_currencies.append(i['symbol'])
        prices = {}

        decimals_dict = {}
        # get the market limits so that we can get accurate rounding
        for i in used_currencies:
            market_limits = self.get_order_filter(i)
            decimals_dict[i] = {}
            decimals_dict[i]['quantity_decimals'] = self.__get_decimals(market_limits['limit_order']['base_increment'])
            decimals_dict[i]['quote_decimals'] = self.__get_decimals(market_limits['market_order']['quote_increment'])

        for i in used_currencies:
            prices[i] = self.get_price(i)
            if not self.backtesting:
                time.sleep(.2)

        for i in range(len(self.paper_trade_orders)):
            index = self.paper_trade_orders[i]
            """
            Coinbase pro example
            {
                "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
                "price": "0.10000000",
                "size": "0.01000000",
                "product_id": "BTC-USD",
                "side": "buy",
                "stp": "dc",
                "type": "limit",
                "time_in_force": "GTC",
                "post_only": false,
                "created_at": "2016-12-08T20:02:28.53864Z",
                "fill_fees": "0.0000000000000000",
                "filled_size": "0.00000000",
                "executed_value": "0.0000000000000000",
                "status": "pending",
                "settled": false
            }
            """
            if index['type'] in ('limit', 'stop_loss') and index['status'] == 'pending':
                current_price = prices[index['symbol']]
                limit_price = index['price']

                if index['side'] == 'buy':
                    if index['price'] > current_price:
                        # Take everything off hold
                        asset_id = index['symbol']
                        quote = utils.get_quote_asset(asset_id)

                        available = self.local_account.get_account(quote)['available']
                        # Put it back into available
                        self.local_account.update_available(quote, available + (index['size'] * index['price']))

                        # Take it out of hold
                        hold = self.local_account.get_account(quote)['hold']
                        self.local_account.update_hold(quote, hold - (index['size'] * index['price']))

                        order, funds, executed_value, fill_fees, filled_size = self.evaluate_paper_trade(index,
                                                                                                         index['price'])
                        self.local_account.trade_local(symbol=index['symbol'],
                                                       side='buy',
                                                       base_delta=filled_size,  # Gain filled size after fees
                                                       quote_delta=funds * -1,  # Loose the original fund amount
                                                       base_resolution=decimals_dict[index['symbol']]
                                                       ['quantity_decimals'],
                                                       quote_resolution=decimals_dict[index['symbol']]
                                                       ['quote_decimals'])
                        order['status'] = 'done'
                        order['settled'] = 'true'

                        # Add this to the executed orders
                        self.executed_orders.append({
                            'id': index['id'],
                            'executed_time': self.time(),
                        })

                        self.paper_trade_orders[i] = order
                elif index['side'] == 'sell':
                    if limit_price < current_price and index['type'] == 'limit' \
                            or current_price <= limit_price and index['type'] == 'stop_loss':
                        # Take everything off hold

                        asset_id = index['symbol']
                        base = utils.get_base_asset(asset_id)

                        available = self.local_account.get_account(base)['available']
                        # Put it back into available
                        self.local_account.update_available(base, available + index['size'])

                        # Remove it from hold
                        hold = self.local_account.get_account(base)['hold']
                        self.local_account.update_hold(base, hold - index['size'])

                        order, funds, executed_value, fill_fees, filled_size = self.evaluate_paper_trade(index,
                                                                                                         index['price'])
                        self.local_account.trade_local(symbol=index['symbol'],
                                                       side='sell',
                                                       base_delta=float(order['size'] * - 1),
                                                       # Loose size before any fees
                                                       quote_delta=executed_value,  # Executed value after fees
                                                       base_resolution=decimals_dict[index['symbol']]
                                                       ['quantity_decimals'],
                                                       quote_resolution=decimals_dict[index['symbol']]['quote_decimals']
                                                       )
                        order['status'] = 'done'
                        order['settled'] = 'true'

                        # Add this to the executed orders
                        self.executed_orders.append({
                            'id': index['id'],
                            'executed_time': self.time(),
                        })

                        self.paper_trade_orders[i] = order

    def evaluate_paper_trade(self, order, current_price):
        """
        This calculates fees & evaluates accurate value
        Args:
            order (dict): Order dictionary to derive the order attributes
            current_price (float): The current price of the currency pair the limit order was created on
        """
        funds = order['size'] * current_price
        executed_value = funds - funds * float((self.get_fees(order['symbol'])["maker_fee_rate"]))
        fill_fees = funds * float((self.get_fees(order['symbol'])["maker_fee_rate"]))
        fill_size = order['size'] - order['size'] * float((self.get_fees(order['symbol'])["maker_fee_rate"]))

        return order, funds, executed_value, fill_fees, fill_size

    @utils.enforce_base_asset
    def get_account(self, symbol=None) -> utils.AttributeDict:
        if symbol is None:
            return self.local_account.get_accounts()
        else:
            # This is a super call in all other interfaces. This apparently doesn't work here because of multiple
            # inheritance
            symbol = utils.get_base_asset(symbol)
            try:
                return self.local_account.get_account(symbol)
            except KeyError:
                if self.backtesting:
                    raise KeyError("Symbol not found. This can be caused by an invalid quote currency "
                                   "in backtest.json.")
                else:
                    raise KeyError("Symbol not found.")

    def take_profit_order(self, symbol: str, price: float, size: float) -> TakeProfitOrder:
        # we don't simulate partial fills, this is the same as take_profit
        order = self.limit_order(symbol, 'sell', price, size)
        return TakeProfitOrder(order._LimitOrder__order, order._LimitOrder__response, order.Interface)

    def stop_loss_order(self, symbol: str, price: float, size: float):
        return self.limit_order(symbol, 'sell', price, size, stop_loss=True)

    def market_order(self, symbol, side, size) -> MarketOrder:
        if not self.backtesting:
            print("Paper Trading...")
        needed = self.needed['market_order']
        creation_time = self.time()
        price = self.get_price(symbol)
        funds = price*size

        market_limits = self.get_order_filter(symbol)

        min_size = market_limits['market_order']["base_min_size"]

        max_size = market_limits['market_order']["base_max_size"]

        if size < min_size:
            raise InvalidOrder(f"Size is too small. Minimum is: {min_size}. You requested {size}.")

        if size > max_size:
            raise InvalidOrder(f"Size is too large. Maximum is: {max_size}. You requested {size}.")

        base_increment = market_limits['market_order']['base_increment']
        base_decimals = self.__get_decimals(base_increment)

        # Test if funds has more decimals than the increment. The increment is the maximum resolution of the quote.
        if self.should_auto_trunc:
            size = utils.trunc(size, base_decimals)
        elif self.__get_decimals(size) > base_decimals:
            raise InvalidOrder("Size resolution is too high, the highest resolution allowed for this symbol is: " +
                               str(base_increment) + ". You specified " + str(size) +
                               ". Try using blankly.trunc(size, decimal_number) to match the exchange resolution.")

        if self.get_exchange_type() == 'alpaca':
            # This could break, but there appears that 10 decimals is about right for alpaca
            quantity_decimals = 10
            shortable = market_limits['exchange_specific']['shortable']
        else:
            shortable = False
            quantity_decimals = self.__get_decimals(market_limits['limit_order']['base_increment'])

        order = {
            'size': size,
            'side': side,
            'symbol': symbol,
            'type': 'market'
        }
        qty = size

        # Test the purchase
        self.local_account.test_trade(symbol, side, qty, price, market_limits['market_order']["quote_increment"],
                                      quantity_decimals,
                                      (shortable and self.__enable_shorting) or self.__force_shorting,
                                      calculate_margin=self.__calculate_margin)
        # Create coinbase pro-like id
        coinbase_pro_id = paper_trade.generate_coinbase_pro_id()
        # TODO the force typing here isn't strictly necessary because its run int the isolate_specific anyway
        response = {
            'symbol': str(symbol),
            'id': str(coinbase_pro_id),
            'created_at': str(creation_time),
            # 'funds': str(utils.trunc(size - size * float((self.__exchange_properties["taker_fee_rate"])),
            #                          base_decimals)),
            'size': str(size),
            'status': 'done',
            'type': 'market',
            'side': str(side),
            # 'specified_funds': str(funds),
            # 'post_only': 'false',
            # 'done_at': str(self.time()),
            # 'done_reason': 'filled',
            # 'fill_fees': str(utils.trunc(funds * float((self.__exchange_properties["maker_fee_rate"])),
            #                              quote_decimals)),
            # 'filled_size': str(utils.trunc(funds - funds *
            #                                float((self.__exchange_properties["maker_fee_rate"])) / price,
            #                                quantity_decimals)),
            # 'executed_value': str(utils.trunc(funds - funds * float((self.__exchange_properties["maker_fee_rate"])),
            #                                   quote_decimals)),
            # 'settled': 'true'
            'exchange_specific': {}
        }
        response = utils.isolate_specific(needed, response)
        self.paper_trade_orders.append(response)
        # Identify the trade also by exchange
        if self.backtesting:
            self.paper_trade_orders[-1]['exchange'] = self.get_exchange_type()

        if side == "buy":
            self.local_account.trade_local(symbol=symbol,
                                           side=side,
                                           base_delta=utils.trunc(qty -
                                                                  qty *
                                                                  float((self.get_fees(symbol)["taker_fee_rate"])),
                                                                  quantity_decimals),
                                           # Gain filled size after fees
                                           quote_delta=utils.trunc(funds * -1,
                                                                   base_decimals),  # Loose the original fund amount
                                           quote_resolution=base_decimals, base_resolution=quantity_decimals)
        elif side == "sell":
            self.local_account.trade_local(symbol=symbol,
                                           side=side,
                                           base_delta=utils.trunc(float(qty * -1), quantity_decimals),
                                           # Loose size before any fees
                                           quote_delta=utils.trunc(funds - funds *
                                                                   float((self.get_fees(symbol)[
                                                                       "taker_fee_rate"])),
                                                                   base_decimals),  # Gain executed value after fees
                                           quote_resolution=base_decimals, base_resolution=quantity_decimals)
        else:
            raise APIException("Invalid trade side: " + str(side))

        self.__check_trading_assets(symbol)

        self.market_order_execution_details.append({
            'id': response['id'],
            'executed_price': price
        })
        return MarketOrder(order, response, self)

    def limit_order(self, symbol, side, price, size, stop_loss=False) -> LimitOrder:
        """
        Used for buying or selling limit orders
        Args:
            symbol: currency to buy
            side: buy/sell
            price: price to set limit order
            size: amount of currency (like BTC) for the limit to be valued
        """
        needed = self.needed['limit_order']
        """
        {
            "id": "d0c5340b-6d6c-49d9-b567-48c4bfca13d2",
            "price": "0.10000000",
            "size": "0.01000000",
            "product_id": "BTC-USD",
            "side": "buy",
            "stp": "dc",
            "type": "limit",
            "time_in_force": "GTC",
            "post_only": false,
            "created_at": "2016-12-08T20:02:28.53864Z",
            "fill_fees": "0.0000000000000000",
            "filled_size": "0.00000000",
            "executed_value": "0.0000000000000000",
            "status": "pending",
            "settled": false
        }
        """
        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'stop_loss' if stop_loss else 'limit'
        }
        creation_time = self.time()

        order_filter = self.get_order_filter(symbol)

        min_base = float(order_filter['limit_order']["base_min_size"])
        max_base = float(order_filter['limit_order']["base_max_size"])

        if size < min_base:
            raise InvalidOrder("Order quantity is too small. Minimum is: " + str(min_base))

        if size > max_base:
            raise InvalidOrder("Order quantity is too large. Maximum is: " + str(max_base))

        min_price = float(order_filter['limit_order']["min_price"])
        max_price = float(order_filter['limit_order']["max_price"])

        if price < min_price:
            raise InvalidOrder("Limit price is too small. Minimum is: " + str(min_price))

        if price > max_price:
            raise InvalidOrder("Limit price is too large. Maximum is: " + str(max_price))

        # Check if the passed parameters are more accurate than either of the max base resolution or max price
        # resolution
        price_increment = order_filter['limit_order']['price_increment']
        price_increment_decimals = self.__get_decimals(price_increment)

        if self.should_auto_trunc:
            price = utils.trunc(price, price_increment_decimals)
        elif self.__get_decimals(price) > price_increment_decimals:
            raise InvalidOrder("Fund resolution is too high, minimum resolution is: " + str(price_increment) +
                               ". Try using blankly.trunc(size, decimal_number) to match the exchange resolution.")

        base_increment = order_filter['limit_order']['base_increment']
        base_decimals = self.__get_decimals(base_increment)

        if self.should_auto_trunc:
            size = utils.trunc(size, base_decimals)
        elif self.__get_decimals(size) > base_decimals:
            raise InvalidOrder("Fund resolution is too high, minimum resolution is: " + str(base_increment) +
                               '. Try using blankly.trunc(size, decimal_number) to match the exchange resolution.')

        order = {
            'size': size,
            'side': side,
            'price': price,
            'symbol': symbol,
            'type': 'limit'
        }

        # Test the trade
        self.local_account.test_trade(symbol, side, size, price, quote_resolution=price_increment_decimals,
                                      base_resolution=base_decimals, shortable=False)

        # Create coinbase pro-like id
        coinbase_pro_id = paper_trade.generate_coinbase_pro_id()
        # TODO the response fills in the exchange specific tags into the base dictionary. This is
        #  useful but really shouldn't be included. We actually really need executed value as a
        #  universally included key.
        response = {
            'symbol': str(symbol),
            'id': str(coinbase_pro_id),
            'created_at': str(creation_time),
            'price': str(price),
            'size': str(size),
            'status': 'pending',
            "time_in_force": "GTC",
            'type': 'stop_loss' if stop_loss else 'limit',
            'side': str(side),
            # 'stp': 'dc',
            # 'post_only': 'false',
            # "fill_fees": "0.0000000000000000",
            # "filled_size": "0.00000000",
            # "executed_value": "0.0000000000000000",
            # 'settled': 'false'
            'exchange_specific': {}
        }
        response = utils.isolate_specific(needed, response)
        self.paper_trade_orders.append(response)
        # Identify the trade also by exchange
        self.paper_trade_orders[-1]['exchange'] = self.get_exchange_type()

        base = utils.get_base_asset(symbol)
        quote = utils.get_quote_asset(symbol)

        if side == "buy":
            available = self.local_account.get_account(quote)['available']
            # Loose the funds when buying
            self.local_account.update_available(quote, available - (size * price))

            # Gain the funds on hold when buying
            hold = self.local_account.get_account(quote)['hold']
            self.local_account.update_hold(quote, hold + (size * price))
        elif side == "sell":
            available = self.local_account.get_account(base)['available']
            # Loose the size when selling
            self.local_account.update_available(base, available - size)

            # Gain size on hold when selling
            hold = self.local_account.get_account(base)['hold']
            self.local_account.update_hold(base, hold + size)
        else:
            raise APIException(f"Invalid side {side}")
        # TODO this is super stinky but refactoring this code is even stinkier
        return StopLossOrder(order, response, self) if stop_loss else LimitOrder(order, response, self)

    def cancel_order(self, symbol, order_id) -> dict:
        """
        This block could potentially work for both exchanges
        """
        del symbol
        order_index = None
        for i in range(len(self.paper_trade_orders)):
            index = self.paper_trade_orders[i]
            if index['status'] == 'pending' and index['id'] == order_id:
                order_index = i
                break

        if order_index is not None:
            # Now that we found it make sure that we move the funds back on available
            order = self.paper_trade_orders[order_index]
            side = order['side']
            size = order['size']
            symbol = order['symbol']
            price = order['price']
            base_asset = utils.get_base_asset(symbol)
            quote_asset = utils.get_quote_asset(symbol)

            if side == 'buy':
                # When you cancel on the buy side you get those funds back in available
                available = self.local_account.get_account(quote_asset)['available']
                self.local_account.update_available(quote_asset, available + (size * price))

                # And loose them on hold
                hold = self.local_account.get_account(quote_asset)['hold']
                self.local_account.update_hold(quote_asset, hold - (size * price))
            elif side == 'sell':
                # Canceling a sell you gain the size back into available
                available = self.local_account.get_account(base_asset)['available']
                self.local_account.update_available(base_asset, available + size)

                # And you loose it in the hold
                hold = self.local_account.get_account(base_asset)['hold']
                self.local_account.update_hold(base_asset, hold - size)

            # This saves the order
            order_id = order['id']
            # Make sure to save this as a canceled order just before closing it
            # Make sure to write in the time also
            self.canceled_orders.append({
                'id': self.paper_trade_orders[order_index]['id'],
                'canceled_time': self.time()
            })

            del self.paper_trade_orders[order_index]
            return {"order_id": order_id}
        else:
            raise APIException("Order ID not found.")

    def get_open_orders(self, symbol=None):
        open_orders = []
        for i in self.paper_trade_orders:
            if i["status"] == "pending":
                open_orders.append(i)
        return open_orders

    def get_order(self, symbol, order_id) -> dict:
        for i in self.paper_trade_orders:
            if i["id"] == order_id:
                return i

    def get_products(self):
        def get_keyless_products():
            symbols_ = []
            for full_symbol in self.full_prices:
                symbols_.append({
                    'symbol': full_symbol
                })
            self.get_products_cache = symbols_
        if self.backtesting:
            if self.get_products_cache is None:
                if self.get_exchange_type() == 'keyless':
                    get_keyless_products()
                else:
                    self.get_products_cache = self.calls.get_products()
            return self.get_products_cache
        else:
            if self.get_exchange_type() == 'keyless':
                get_keyless_products()
                return self.get_products_cache
            else:
                return self.calls.get_products()

    def get_fees(self, symbol):
        if self.backtesting:
            # Add exchanges that actually require a symbol
            type_ = self.get_exchange_type()
            if type_ == 'okx' or type_ == 'binance':
                if symbol not in self.get_fees_cache:
                    self.get_fees_cache[symbol] = self.calls.get_fees(symbol)
                    return self.get_fees_cache[symbol]
                else:
                    return self.get_fees_cache[symbol]
            # If it doesn't require a symbol just store it in the root
            else:
                if self.get_fees_cache == {}:
                    self.get_fees_cache = self.calls.get_fees(symbol)
                    return self.get_fees_cache
                else:
                    return self.get_fees_cache
        else:
            return self.calls.get_fees(symbol)

    def get_product_history(self, symbol, epoch_start, epoch_stop, resolution):
        if self.backtesting:
            return utils.extract_price_by_resolution(self.full_prices, symbol, epoch_start, epoch_stop, resolution)
        else:
            return self.calls.get_product_history(symbol, epoch_start, epoch_stop, resolution)

    def get_order_filter(self, symbol):
        # Don't re-query order filter if its cached
        if symbol not in self.get_order_filter_cache:
            self.get_order_filter_cache[symbol] = self.calls.get_order_filter(symbol)

        if self.get_exchange_type() == 'binance':
            self.get_order_filter_cache[symbol] = self.__evaluate_binance_limits(self.get_price(symbol),
                                                                                 self.get_order_filter_cache[symbol])
        return self.get_order_filter_cache[symbol]

    def get_price(self, symbol) -> float:
        if self.backtesting:
            return self.get_backtesting_price(symbol)
        else:
            def idle(tick):
                pass
            # Only get crazy websocket prices if asked
            if self.user_preferences['settings']['paper']['price_source'] == 'websocket':
                tickers = self.__ticker_manager.get_all_tickers()
                if self.get_exchange_type() in tickers and symbol in tickers[self.get_exchange_type()]:
                    most_recent_tick = tickers[self.get_exchange_type()][symbol].get_most_recent_tick()

                    if most_recent_tick is None:
                        utils.info_print("No data found on ticker yet - using API...")
                        return self.calls.get_price(symbol)
                    else:
                        return most_recent_tick['price']
                else:
                    utils.info_print(f"Creating ticker on symbol {symbol} for exchange {self.get_exchange_type()}...")
                    self.__ticker_manager.create_ticker(callback=self._websocket_update, override_symbol=symbol,
                                                        override_exchange=self.get_exchange_type())
                    return self.calls.get_price(symbol)
            else:
                return self.calls.get_price(symbol)

    @staticmethod
    def __evaluate_binance_limits(price: (int, float), order_filter):
        order_filter['limit_order']['min_price'] = order_filter['exchange_specific']['limit_multiplier_down'] * price
        order_filter['limit_order']['max_price'] = order_filter['exchange_specific']['limit_multiplier_up'] * price
        return order_filter



from blankly.frameworks.strategy.strategy_base import StrategyBase
from blankly.frameworks.strategy.strategy_state import StrategyState
from blankly.frameworks.strategy.strategy import Strategy
from blankly.frameworks.strategy.futures_strategy import FuturesStrategy
from blankly.frameworks.strategy.futures_strategy_state import FuturesStrategyState



"""
    OANDA root API calls & definitions.
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import json

import requests
from collections import OrderedDict

from blankly.utils.exceptions import APIException


def api_error_handler(func):
    def wrapper(*args, **kwargs):
        resp = func(*args, **kwargs)
        if 'errorMessage' in resp:
            raise APIException(str(resp))
        return resp
    return wrapper


class OandaAPI:
    # documentation here: http://developer.oanda.com/rest-live-v20/account-ep/
    API_URL = 'https://api-fxtrade.oanda.com'
    API_PRACTICE_URL = 'https://api-fxpractice.oanda.com'

    def __init__(self, personal_access_token: str, account_id: str, sandbox: bool = False):
        self.__api_key = personal_access_token
        self.default_account = account_id

        if not sandbox:
            self.__api_url = self.API_URL
        else:
            self.__api_url = self.API_PRACTICE_URL

        self.session = self._init_session()

    def _init_session(self):
        session = requests.session()
        session.headers.update({"Content-Type": "application/json",
                                "Accept-Datetime-Format": "UNIX",
                                'Authorization': 'Bearer {}'.format(self.__api_key)})
        return session

    def _send_request(self, method, url, params=None, data=None):
        if not params:
            params = OrderedDict()
        else:
            assert isinstance(params, OrderedDict)

        if not data:
            data = OrderedDict()
        else:
            assert isinstance(data, OrderedDict)

        response = getattr(self.session, method)(url, params=params, data=json.dumps(data))
        return response.json()

    """
    Account Endpoints
    """

    def get_all_accounts(self):
        endpoint = '/v3/accounts'
        return self._send_request('get', self.__api_url + endpoint)

    def get_account(self, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)

        endpoint = f'/v3/accounts/{accountid}'
        return self._send_request('get', self.__api_url + endpoint)

    def get_account_summary(self, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/summary'
        return self._send_request('get', self.__api_url + endpoint)

    def get_account_instruments(self, instrument: str = None, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/instruments'

        params = OrderedDict()
        if instrument is not None:
            params["instruments"] = instrument

        return self._send_request('get', self.__api_url + endpoint, params)

    def get_account_changes(self, sinceTransactionID: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/changes'

        params = OrderedDict()
        params["sinceTransactionID"] = sinceTransactionID

        return self._send_request('get', self.__api_url + endpoint, params=params)

    def patch_account_configuration(self, alias: str, marginRate: float, accountid: str):
        # Client-defined alias (name) for the Account
        # alias: (string),
        # The string representation of a decimal number.
        # marginRate: (DecimalNumber)
        if accountid is None:
            accountid = self.default_account

        assert isinstance(accountid, str)
        assert isinstance(alias, str)
        assert isinstance(marginRate, float)
        endpoint = f'/v3/accounts/{accountid}/configuration'

        data = OrderedDict()
        data["alias"] = alias
        data["marginRate"] = marginRate

        return self._send_request('patch', self.__api_url + endpoint, data=data)

    """
    Instrument Endpoints
    """

    def get_candles_by_startend(self, instrument: str, granularity: str, from_unix: float, to_unix: float):
        endpoint = f'/v3/instruments/{instrument}/candles'
        params = OrderedDict()
        params["granularity"] = granularity
        params["from"] = str(from_unix)
        params["to"] = str(to_unix)

        return self._send_request('get', self.__api_url + endpoint, params=params)

    def get_last_k_candles(self, instrument: str, granularity: str, to_unix: [int, float], count: int):
        endpoint = f'/v3/instruments/{instrument}/candles'
        params = OrderedDict()
        params["granularity"] = granularity
        params["count"] = count
        params["to"] = str(to_unix)

        return self._send_request('get', self.__api_url + endpoint, params=params)

    def get_order_book(self, instrument: str):
        assert isinstance(instrument, str)
        endpoint = f'/v3/instruments/{instrument}/orderBook'
        return self._send_request('get', self.__api_url + endpoint)

    def get_position_book(self, instrument: str):
        assert isinstance(instrument, str)
        endpoint = f'/v3/instruments/{instrument}/positionBook'
        return self._send_request('get', self.__api_url + endpoint)

    """
    Order Endpoints
    """

    def get_orders(self, instrument: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders'

        params = OrderedDict()
        params["instrument"] = instrument
        params["count"] = 500

        return self._send_request('get', self.__api_url + endpoint,  params=params)

    def get_all_open_orders(self, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/pendingOrders'
        return self._send_request('get', self.__api_url + endpoint)

    @api_error_handler
    def get_order(self, orderid: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders/{orderid}'
        return self._send_request('get', self.__api_url + endpoint)

    @api_error_handler
    def cancel_order(self, orderid: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders/{orderid}/cancel'
        return self._send_request('put', self.__api_url + endpoint)

    def place_market_order(self, instrument: str, units: float, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders'

        order_request = {
            "type": "MARKET",
            "instrument": instrument,
            "units": units,
            "timeInForce": "FOK"
        }
        data = OrderedDict()
        data["order"] = order_request

        return self._send_request('post', self.__api_url + endpoint, data=data)

    def place_limit_order(self, instrument: str, units: float, price: float, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders'

        order_request = {
            "type": "LIMIT",
            "instrument": instrument,
            "units": units,
            "price": price,
            "timeInForce": "GTC"
        }
        data = OrderedDict()
        data["order"] = order_request

        return self._send_request('post', self.__api_url + endpoint, data=data)

    def place_stop_loss(self, instrument: str, units: float, price: float, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders'

        order_request = {
            "type": "STOP",
            "instrument": instrument,
            "units": units,
            "price": price,
            "timeInForce": "GTC"
        }
        data = OrderedDict()
        data["order"] = order_request

        return self._send_request('post', self.__api_url + endpoint, data=data)

    def place_take_profit(self, instrument: str, units: float, price: float, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/orders'

        order_request = {
            "type": "TAKE_PROFIT",
            "instrument": instrument,
            "units": units,
            "price": price,
            "timeInForce": "GTC"
        }
        data = OrderedDict()
        data["order"] = order_request

        return self._send_request('post', self.__api_url + endpoint, data=data)

    """
    Trade Endpoints
    """

    """
    Position Endpoints
    """

    def get_all_positions(self, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/positions'
        return self._send_request('get', self.__api_url + endpoint)

    def get_all_open_positions(self, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/openPositions'
        return self._send_request('get', self.__api_url + endpoint)

    def get_position(self, instrument: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/positions/{instrument}'
        return self._send_request('get', self.__api_url + endpoint)

    def close_position(self, instrument: str, accountid: str = None):
        if accountid is None:
            accountid = self.default_account
        assert isinstance(accountid, str)
        endpoint = f'/v3/accounts/{accountid}/positions/{instrument}/close'
        return self._send_request('put', self.__api_url + endpoint)

    """
    Transaction Endpoints
    """

    """
    Pricing Endpoints
    """



"""
    Logic to provide consistency across exchanges
    Copyright (C) 2021  Emerson Dove

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import abc
from functools import lru_cache

import blankly.utils.utils as utils
from blankly.exchanges.interfaces.abc_exchange_interface import ABCExchangeInterface


# TODO: need to add a cancel all orders function
class ExchangeInterface(ABCExchangeInterface, abc.ABC):
    def __init__(self, exchange_name, authenticated_api, preferences_path=None, valid_resolutions=None):
        self.exchange_name = exchange_name
        self.calls = authenticated_api
        self.valid_resolutions = valid_resolutions
        # Reload user preferences here
        self.user_preferences = utils.load_user_preferences(preferences_path)

        self.exchange_properties = None
        # Some exchanges like binance will not return a value of 0.00 if there is no balance
        self.available_currencies = {}

        self.needed = {
            '__init_exchange__': [
                ['maker_fee_rate', float],
                ['taker_fee_rate', float]
            ],
            'get_products': [
                ["symbol", str],
                ["base_asset", str],
                ["quote_asset", str],
                ["base_min_size", float],
                ["base_max_size", float],
                ["base_increment", float]
            ],
            'get_account': [
                ["available", float],
                ["hold", float]
            ],
            'market_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["size", float],
                ["status", str],
                ["type", str],
                ["side", str]
            ],
            'limit_order': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["price", float],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'take_profit': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["stop_price", float],
                ["stop", str],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'stop_loss': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["stop_price", float],
                ["stop", str],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'stop_limit': [
                ["symbol", str],
                ["id", str],
                ["created_at", float],
                ["stop_price", float],
                ["limit_price", float],
                ["stop", str],
                ["size", float],
                ["status", str],
                ["time_in_force", str],
                ["type", str],
                ["side", str]
            ],
            'cancel_order': [
                ['order_id', str]
            ],
            # 'get_open_orders': [  # Key specificity changes based on order type
            #     ["id", str],
            #     ["price", float],
            #     ["size", float],
            #     ["type", str],
            #     ["side", str],
            #     ["status", str],
            #     ["product_id", str]
            # ],
            # 'get_order': [
            #     ["product_id", str],
            #     ["id", str],
            #     ["price", float],
            #     ["size", float],
            #     ["type", str],
            #     ["side", str],
            #     ["status", str],
            #     ["funds", float]
            # ],
            'get_fees': [
                ['maker_fee_rate', float],
                ['taker_fee_rate', float]
            ],
            'get_order_filter': [
                ["symbol", str],
                ["base_asset", str],
                ["quote_asset", str],
                ["max_orders", int],
                ["limit_order", dict],
                ["market_order", dict]
            ]
        }

        if self.user_preferences['settings']['test_connectivity_on_auth']:
            self.init_exchange()

    @abc.abstractmethod
    def init_exchange(self):
        """
        Create the properties for the exchange.

        This is never run if test_connectivity_on_auth is set to false
        """
        pass

    """ Needs to be overridden here """

    def get_calls(self):
        """
        Returns:
             The exchange's direct calls object. A blankly Bot class should have immediate access to this by
             default
        """
        return self.calls

    """ Needs to be overridden here """

    def get_exchange_type(self):
        return self.exchange_name

    @property
    def account(self):
        return utils.AttributeDict(self.get_account())

    @property
    def orders(self):
        return self.get_open_orders()

    @property
    def cash(self):
        using_setting = self.user_preferences['settings'][self.exchange_name]['cash']
        return self.get_account(using_setting)['available']

    def backtesting_time(self):
        """
        Overridden by interfaces which have a valid backtesting boolean value
        """
        return None

    @staticmethod
    def evaluate_multiples(valid_resolutions: list, resolution_seconds: float):
        found_multiple = -1
        for multiple in reversed(valid_resolutions):
            if resolution_seconds % multiple == 0:
                found_multiple = multiple
                break
        if found_multiple < 0:
            raise ValueError("This exchange currently does not support this specific resolution, try making it a "
                             "multiple of 1 minute, 1 hour or 1 day")

        row_divisor = resolution_seconds / found_multiple
        row_divisor = int(row_divisor)

        if row_divisor > 100:
            raise Warning("The resolution you requested is an extremely high of the base resolutions supported and may "
                          "slow down the performance of your model: {} * {}".format(found_multiple, row_divisor))

        return found_multiple, row_divisor

    def get_account(self, symbol=None):
        """
        This method is a helper that allows the get_account functions to assume the base asset is being passed in
        Args:
            symbol (Optional): Filter by particular symbol

        Coinbase Pro: get_account
        binance: get_account["balances"]
        """

        if symbol is not None:
            symbol = utils.get_base_asset(symbol)

        return symbol

    def choose_order_specificity(self, order_type):
        # This lower should not be necessary if everything is truly homogeneous
        order_type = order_type.lower()
        if order_type == 'stop':
            return self.needed['stop_loss']
        elif order_type == 'take_profit':
            return self.needed['take_profit']
        return self.needed[f'{order_type}_order']

    """
    Order lifecycle should be:
    Accepted -> live -> done -> settled
    """

    @staticmethod
    def homogenize_order_status(exchange, status):
        if exchange == "binance":
            if status == "new":
                return "open"
        elif exchange == 'alpaca':
            if status == 'accepted':
                return 'accepted'
            if status == 'new':
                return 'new'

        return status

    @property
    def should_auto_trunc(self):
        return self.user_preferences['settings'].get('auto_truncate', False)

    @lru_cache(None)
    def get_asset_precision(self, asset):
        try:
            product = next(p for p in self.get_products() if p['symbol'] == asset)
            return utils.increment_to_precision(product['base_increment'])
        except (KeyError, StopIteration):
            return 8  # reasonable default for symbols that don't exist



"""
    Functions for blankly login/logout commands
    Copyright (C) 2022 Matias Kotlik

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import json
import pathlib
import sys
import urllib.parse
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Optional

from blankly.deployment.api import API
from blankly.deployment.ui import print_work, print_failure, print_success, show_spinner

SUCCESS_URL = 'https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/login_success.html?alt=media&token=41d734e2-0a88-44c4-b1dd-7e081fd019e7'
FAILURE_URL = 'https://firebasestorage.googleapis.com/v0/b/blankly-6ada5.appspot.com/o/login_failure.html?alt=media&token=b2d9f6dc-aa54-4da2-aa6f-1f75a4025634'


def write_token(token):
    token_file = get_token_file()
    token_file.parent.mkdir(parents=True, exist_ok=True)
    with open(token_file, 'w') as file:
        json.dump({'token': token}, file)


def poll_login() -> Optional[API]:
    # mutate a list to avoid needing global variable shit and piss
    token = []

    class TokenListener(BaseHTTPRequestHandler):
        def do_GET(self):
            try:
                token.append(urllib.parse.parse_qs(self.path[2:])['token'][0])
                self.send_response_only(302)
                self.send_header('Location', SUCCESS_URL)
                self.end_headers()
            except Exception:
                token.append(False)
                self.send_response(302)
                self.send_header('Location', FAILURE_URL)  # send them back lmao
                self.end_headers()

    server = HTTPServer(('', 9082), TokenListener)

    while not token:
        server.handle_request()
    token = token[0]

    if token:
        api = API(token)
        write_token(token)
        return api


def get_token() -> Optional[str]:
    token_file = get_token_file()
    if token_file.exists():
        # file exists, parse it or error out
        with open(token_file, 'r') as file:
            data = json.load(file)
        return data['token']
    return None


def get_token_file() -> pathlib.Path:
    return get_datadir() / 'blankly' / 'auth.json'


# https://stackoverflow.com/a/61901696
def get_datadir() -> pathlib.Path:
    """
    Returns a parent directory path where persistent application data can be stored.

    # linux: ~/.local/share
    # macOS: ~/Library/Application Support
    # windows: C:/Users/<USER>/AppData/Roaming
    """

    home = pathlib.Path.home()

    if sys.platform == "win32":
        return home / "AppData/Roaming"
    elif sys.platform == "darwin":
        return home / "Library/Application Support"
    elif sys.platform.startswith("linux"):
        return home / ".local/share"
    return home


def logout():
    # python 3.8 has a parameter for this `missing_ok` but we need to support 3.7
    try:
        get_token_file().unlink()
    except FileNotFoundError:
        pass
